#pragma once

#include "vulkan_allocator.hpp"
#include "dynamic_size_block_allocator.hpp"
#include "vertex.hpp"

#include <gz-util/concepts.hpp>

#include <cstdint>

namespace gz::vlk {
	constexpr size_t TODO_VERTICES_COUNT = 100000;
	constexpr size_t TODO_INDICES_COUNT = 100000;
	constexpr uint32_t BUFFER_NOT_INITIALIZED = UINT32_MAX;
	struct BufferInfo {
		uint32_t index = BUFFER_NOT_INITIALIZED;
		uint32_t offset = BUFFER_NOT_INITIALIZED;
		uint32_t count = BUFFER_NOT_INITIALIZED;
		std::string toString() const;
	};

	struct Buffer {
		vk::Buffer buffer;
		MemoryInfo bufferMemory;
		vk::DeviceSize bufferSize;
		DynBlockAllocator allocator;
	};

	// defined in vulkan_instance.hpp
	class VulkanInstance;

	/**
	 * @brief Manage multiple vertex and index buffers on device local memory
	 * @details
	 *  Instantiated for Vertex2D, Vertex3D with uint16_t and uint32_t respectively.
	 *  Manages buffers allocated from VulkanAllocator. 
	 *  The vertex and index buffers are then suballocated from those buffers (using DynBlockAllocator).
	 */
	template<VertexType VertexT, SupportedIndexType IndexT>
	class BufferManager {
		public:
			BufferManager(VulkanInstance& instance);

			vk::Buffer getVertexBuffer(const BufferInfo& vertexBufferInfo) const;
			vk::Buffer getIndexBuffer(const BufferInfo& indexBufferInfo) const;

			/**
			 * @brief Copy vertices and indices to a gpu-local buffer
			 * @details
			 *  Instantiated for std::vectors
			 */
			template<util::ContiguousRange<VertexT> VertexRange, util::ContiguousRange<IndexT> IndexRange>
			[[ nodiscard ]] std::pair<BufferInfo, BufferInfo> addVertices(const VertexRange& vertices, const IndexRange& indices);
			/**
			 * @brief Free the vertex and index buffers and reset bufferInfos
			 * @todo Deallocate memory if unused
			 */
			void removeVertices(BufferInfo& vertexBufferInfo, BufferInfo& indexBufferInfo);
		private:
			/**
			 * @brief Deallocate all buffers
			 * @details
			 *  Called by VulkanInstance::cleanup()
			 */
			void cleanup();
			std::vector<Buffer> vertexBuffers;
			std::vector<Buffer> indexBuffers;

			VulkanInstance& vk;
			Log bLog;
	};

}  // namespace gz::vlk
