#include "renderer2D.hpp"

#include "exceptions.hpp"
#include "shape.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_instance.hpp"
#include "texture_manager.hpp"

#include <cstdint>
#include <cstring>
#include <stdexcept>
#include <thread>
#include <vulkan/vulkan_enums.hpp>

namespace gz::vlk {
#define RENDERER_2D_RENDER_TO_OWN_IMAGES
#define RENDERER_2D_COPY_TO_SWAP_CHAIN

#ifdef RENDERER_2D_RENDER_TO_SWAP_CHAIN
    #ifdef RENDERER_2D_RENDER_TO_OWN_IMAGES
        static_assert(false, "RENDERER_2D_RENDER_TO_SWAP_CHAIN and RENDERER_2D_RENDER_TO_OWN_IMAGES defined at the same time");
    #endif
    #ifdef RENDERER_2D_COPY_TO_SWAP_CHAIN
        static_assert(false, "RENDERER_2D_RENDER_TO_SWAP_CHAIN and RENDERER_2D_COPY_TO_SWAP_CHAIN defined at the same time");
    #endif
    #ifdef RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        static_assert(false, "RENDERER_2D_RENDER_TO_SWAP_CHAIN and RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES defined at the same time");
    #endif
#elif defined RENDERER_2D_RENDER_TO_OWN_IMAGES
/*     #ifndef RENDERER_2D_COPY_TO_SWAP_CHAIN */
/*         #define RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES */
/*     #endif */
#else
    static_assert(false, "Either RENDERER_2D_RENDER_TO_OWN_IMAGES or RENDERER_2D_RENDER_TO_SWAP_CHAIN must be defined");
#endif

//
// INIT & CLEANUP
//
    Renderer2D::Renderer2D(VulkanInstance& instance, TextureManager& textureManager)
        : Renderer(instance, textureManager), shapeBufferManager(instance)
    {
        LogCreateInfo logCI {
        	.logfile = "renderer2D.log",
        	.storeLog = false,
        	.prefix = "2D-Renderer",
        	.prefixColor = Color::LI_MAGENTA,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        rLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&Renderer2D::cleanup, this));
        vk.registerSwapChainRecreateCallback(std::bind(&Renderer2D::swapChainRecreateCallback, this));
        vk.createCommandBuffers(commandBuffers);
        const size_t vertexCount = 500;
        const size_t indexCount = 1000;
        vk.createVertexBuffer<Vertex2D>(vertexCount, vertexBuffer, vertexBufferMemory, vertexBufferSize);
        vk.createIndexBuffer<uint32_t>(indexCount, indexBuffer, indexBufferMemory, indexBufferSize);
#ifdef RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        createDescriptorResources();
#endif
        initSwapChainDependantResources();
        VulkanInstance::registerObjectUsingVulkan("Renderer2D",
                &vertexBuffer, &vertexBufferMemory.memory, 
                &indexBuffer, &indexBufferMemory.memory,
                &renderPass, &pipelines[R2Render].pipeline, //&pipelines[blendToSC].pipeline,
                &imageSampler, &framebuffers, &renderImages, &renderImageViews,  // missing: &renderImageMemory, 
                &commandBuffers);
        rLog("Created Renderer2D");
    }


    void Renderer2D::cleanup() {
        rLog.log0("cleanup: cleaning up");
        // UPDATE DOC ON CHANGES!

        /* vk.destroyCommandBuffers(commandBuffers); */
#ifdef RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        vk.getDevice().destroyDescriptorPool(descriptorPool);
        vk.getDevice().destroyDescriptorSetLayout(sampleFromRenderImageDSL);
#endif
        cleanupSwapChainDependantResources();
        cleanup_();
    }


//
// SWAPCHAIN DEPENDANT
//
    void Renderer2D::initSwapChainDependantResources() {
        createRenderPass();
        createImageResources();
#ifdef RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        createDescriptorSetSampler();
#endif
        vk.createFramebuffers(framebuffers, renderImageViews, renderPass);
        std::vector<vk::DescriptorSetLayout> descriptorSetLayouts = { textureManager.getDescriptorSetLayout() };

        // specialization constant for sampler2D array length = atlas count
        vk::SpecializationMapEntry specME {
            .constantID = 0,
            .offset = 0,
            .size = sizeof(uint32_t),
        };
        uint32_t atlasCount = textureManager.getAtlasCount();
        vk::SpecializationInfo specI;
        specI.setData<uint32_t>(atlasCount);
        specI.setMapEntries(specME);

        // VERTEX 2D PIPELINE
        auto vertShaderCode = readBinaryFile("shaders/shader2D.vert.spv");
        auto fragShaderCode = readBinaryFile("shaders/shader2D.frag.spv");

        vk::ShaderModule vertShaderModule = vk.createShaderModule(vertShaderCode);
        vk::ShaderModule fragShaderModule = vk.createShaderModule(fragShaderCode);

        std::vector<vk::PipelineShaderStageCreateInfo> shaderStages;

        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eVertex,
        	.module = vertShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = nullptr,
        });
        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eFragment,
        	.module = fragShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = &specI,
        });

        vk::GraphicsPipelineCreateInfo pipelineCI {
            .renderPass = renderPass,
            .subpass = 0,
        };
        pipelineCI.setStages(shaderStages);

        std::vector<vk::PushConstantRange> pushConstantRanges;
        pushConstantRanges.emplace_back(vk::PushConstantRange{
            .stageFlags = vk::ShaderStageFlagBits::eFragment,
            .size = sizeof(uint32_t),
        });
        vk.createGraphicsPipeline<Vertex2D>(std::move(pipelineCI), descriptorSetLayouts, pushConstantRanges, false, pipelines[R2Render]);

        vk.getDevice().destroyShaderModule(vertShaderModule, nullptr);
        vk.getDevice().destroyShaderModule(fragShaderModule, nullptr);
    }


    void Renderer2D::cleanupSwapChainDependantResources() {
        // destroy pipelines
        pipelines.clear(vk.getDevice());

        vk.destroyFramebuffers(framebuffers);

#ifdef RENDERER_2D_RENDER_TO_OWN_IMAGES
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.destroyImageView(renderImageViews[i]);
            vk.destroyImage(renderImages[i], renderImageMemory[i]);
        }
#endif
        vk.getDevice().destroyRenderPass(renderPass, nullptr);
    }


    void Renderer2D::swapChainRecreateCallback() {
        cleanupSwapChainDependantResources();
        initSwapChainDependantResources();
    }


//
// IMAGES
//
    void Renderer2D::createImageResources() {
        renderImageViews.resize(vk.getScImages().size());
#ifdef RENDERER_2D_RENDER_TO_OWN_IMAGES
        renderImages.resize(vk.getScImages().size());
        renderImageMemory.resize(vk.getScImages().size());
        vk::ImageUsageFlags usage = vk::ImageUsageFlagBits::eColorAttachment;
    #ifdef RENDERER_2D_SAMPLE_FROM_RENDER_IMAGES
        usage |= vk::ImageUsageFlagBits::eSampled;
    #elif defined RENDERER_2D_COPY_TO_SWAP_CHAIN
        usage |= vk::ImageUsageFlagBits::eTransferSrc;
    #endif
        rLog.log("Image usage flags:", usage);
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImage(vk.getScExtent().width, vk.getScExtent().height, vk.getScImageFormat(), vk::ImageTiling::eOptimal, usage, vk::MemoryPropertyFlagBits::eDeviceLocal, renderImages[i], renderImageMemory[i]);
        }
    #ifdef RENDERER_2D_SAMPLE_FROM_RENDER_IMAGES
        vk.createTextureSampler(imageSampler);
    #endif
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImageView(vk.getScImageFormat(), renderImages[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        rLog.log0("createImageResources: created images, views and sampler");
#elif defined RENDERER_2D_RENDER_TO_SWAP_CHAIN
        // views to swap chain images
        for (size_t i = 0; i < renderImageViews.size(); i++) {
            vk.createImageView(vk.getScImageFormat(), vk.getScImages()[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        rLog.log0("createImageResources: created image views");
#endif
        assert(renderImageViews.size() == vk.getScImages().size());
    }

//
// RENDER PASS
//
    void Renderer2D::createRenderPass() {
        vk::AttachmentDescription2 colorBlendAttachment {
            .format = vk.getScImageFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eStore,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
#ifdef RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
            .finalLayout = vk::ImageLayout::eShaderReadOnlyOptimal,
#elif defined RENDERER_2D_COPY_TO_SWAP_CHAIN
            .finalLayout = vk::ImageLayout::eTransferSrcOptimal,
#else
            .finalLayout = vk::ImageLayout::eColorAttachmentOptimal,
#endif
        };

        vk::AttachmentReference2 colorAttachmentRef {
            .attachment = 0,
            .layout = vk::ImageLayout::eColorAttachmentOptimal,
        };

        // SUBPASS: Rendering
        vk::SubpassDescription2 drawSubpass {
            .pipelineBindPoint = vk::PipelineBindPoint::eGraphics,
            .colorAttachmentCount = 1,
            .pColorAttachments = &colorAttachmentRef,
            /* .pDepthStencilAttachment = &depthAttachmentRef, */
        };

        vk::SubpassDependency2 dsColorAttachmentSD {
            .srcSubpass = VK_SUBPASS_EXTERNAL,
            .dstSubpass = 0,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .srcAccessMask = NO_ACC_FLAGS,
            .dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
        };

        // dependecy for the image layout transition to transfer dst
        vk::SubpassDependency2 dsLayoutTransitionSD {
            .srcSubpass = 0,
            .dstSubpass = VK_SUBPASS_EXTERNAL,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eTransfer,
            .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
            .dstAccessMask = vk::AccessFlagBits::eTransferRead,
            .dependencyFlags = vk::DependencyFlagBits::eByRegion,
        };

        // SUBPASS 2: BLIT ONTO SWAP CHAIN IMAGE
        /* vk::SubpassDescription2 blitSubpass { */
        /*     .pipelineBindPoint = vk::PipelineBindPoint::eGraphics, */
        /*     .colorAttachmentCount = 1, */
        /*     .pColorAttachments = &colorAttachmentRef, */
        /* }; */

        /* vk::SubpassDependency2 bsColorAttachmentSD { */
        /*     .srcSubpass = 0, */
        /*     .dstSubpass = 1, */
        /*     .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput, */
        /*     .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput, */
        /*     .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite, */
        /*     .dstAccessMask = vk::AccessFlagBits::eColorAttachmentRead, */
        /* }; */

        std::vector<vk::AttachmentDescription2> attachments = { colorBlendAttachment };
        std::vector<vk::SubpassDescription2> subpasses { drawSubpass };
        std::vector<vk::SubpassDependency2> dependencies = { dsColorAttachmentSD, dsLayoutTransitionSD };  //, bsColorAttachmentSD }; //dsLayoutTransitionSD };
        vk::RenderPassCreateInfo2 renderPassCI;
        renderPassCI.setAttachments(attachments);
        renderPassCI.setSubpasses(subpasses);
        renderPassCI.setDependencies(dependencies);

        vk::Result result = vk.getDevice().createRenderPass2(&renderPassCI, nullptr, &renderPass); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Could not create render pass", "Renderer2D::createRenderPass");
        }
        rLog("createRenderPass: Created render pass.");
    }

//
// DESCRIPTORS
//
    void Renderer2D::createDescriptorResources() {
        createDescriptorSetLayoutSampler(0);
        // POOL
        std::vector<vk::DescriptorPoolSize> poolSizes;
        poolSizes = { 
            { .type = vk::DescriptorType::eCombinedImageSampler, .descriptorCount = static_cast<uint32_t>(vk.getScImages().size()) },
        };
        uint32_t maxSets = 0;
        for (auto& size : poolSizes) {
            maxSets += size.descriptorCount;
        }
        vk.createDescriptorPool(poolSizes, maxSets, descriptorPool);
        rLog("createDescriptorResources: pool", reinterpret_cast<uint64_t>(static_cast<VkDescriptorPool>(descriptorPool)), "vs hpp:", reinterpret_cast<uint64_t>(&(*descriptorPool)));
    }

//
// RENDERING
//
    void Renderer2D::beginFrameDraw(uint32_t imageIndex, uint32_t currentFrame) {
        commandBuffers[currentFrame].reset();
        vk::CommandBufferBeginInfo commandBufferBI {
        	/* .flags = 0, */
        	/* .pInheritanceInfo = nullptr, */
        };
        vk::Result result = commandBuffers[currentFrame].begin(&commandBufferBI);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to begin 2D command buffer", "Renderer2D::beginFrameDraw");
        }

        // clear
        std::array<vk::ClearValue, 1> clearValues{};
        vk::ClearColorValue clearColor{};
        clearColor.setFloat32({ 0.0f, 0.0f, 0.1f, 0.5f });
        clearValues[0].setColor(std::move(clearColor));
        /* clearValues[1].depthStencil = {1.0f, 0}; */
        vk::RenderPassBeginInfo renderPassBI {
        	.renderPass = renderPass,
        	.framebuffer = framebuffers[imageIndex],
            .renderArea { .offset = { 0, 0 }, .extent = vk.getScExtent(), }
        };
        renderPassBI.setClearValues(clearValues);

        commandBuffers[currentFrame].beginRenderPass(&renderPassBI, vk::SubpassContents::eInline);
    }

    
	void Renderer2D::endFrameDraw(uint32_t imageIndex, uint32_t currentFrame) {
        commandBuffers[currentFrame].endRenderPass();

#ifdef RENDERER_2D_COPY_TO_SWAP_CHAIN
        vk.copyImageToImage(commandBuffers[currentFrame], renderImages[imageIndex], vk.getScImages()[imageIndex], vk.getScExtent());
#endif
        vk::Result result = commandBuffers[currentFrame].end();
        if (result != vk::Result::eSuccess) {
            rLog.error("Failed to record 2D - command buffer", "VkResult:", result);
            throw getVkException(result, "Failed to record 2D - command buffer", "Renderer2D::endFrameDraw");
        }
        vk.submitThisFrame(commandBuffers[currentFrame]);
    }


    /* void Renderer2D::fillVertexBufferWithShapes() { */
    /*     rLog("fillVertexBufferWithShapes"); */
    /*     if (vertexBufferSize < shapesVerticesCount * sizeof(Vertex2D)) { */
    /*         throw VkException("vertex buffer too small! vertexBufferSize: " + std::to_string(vertexBufferSize) + ", required size: " + std::to_string(shapesVerticesCount), "fillVertexBufferWithShapes"); */
    /*     } */

    /*     // create staging buffer */
    /*     vk::Buffer stagingBuffer; */
    /*     MemoryInfo stagingBufferMemory; */
    /*     vk.createBuffer(vertexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory); */

    /*     // fill staging buffer */
    /*     void* data; */
    /*     vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, vertexBufferSize, NO_MEM_FLAGS, &data); */
    /*     if (result != vk::Result::eSuccess) { */
    /*         throw getVkException(result, "Failed to map staging buffer", "Renderer2D::fillVertexBufferWithShapes"); */
    /*     } */
    /*     Vertex2D* vdata = reinterpret_cast<Vertex2D*>(data); */
    /*     size_t offset = 0; */
    /*     for (auto it = shapes.begin(); it != shapes.end(); it++) { */
    /*         rLog.log0("fillVertexBufferWithShapes: copying vertex buffer nr", it - shapes.begin(), "-", it->getVertices(), "to address:", long(vdata + offset), " offset:", offset); */
    /*         memcpy(vdata+offset, it->getVertices().data(), it->getVertices().size() * sizeof(Vertex2D)); */
    /*         offset += it->getVertices().size(); */
    /*     } */
    /*     vk.getDevice().unmapMemory(stagingBufferMemory.memory); */
    /*     // fill vertex buffer */ 
    /*     vk.copyBuffer(stagingBuffer, vertexBuffer, vertexBufferSize); */
    /*     vk.destroyBuffer(stagingBuffer, stagingBufferMemory); */
    /* } */
    /* void Renderer2D::fillIndexBufferWithShapes() { */
    /*     rLog("fillIndexBufferWithShapes"); */
    /*     if (indexBufferSize < shapesIndicesCount * sizeof(uint32_t)) { */
    /*         throw VkException("index buffer too small! indexBufferSize: " + std::to_string(vertexBufferSize) + ", required size: " + std::to_string(shapesVerticesCount), "fillVertexBufferWithShapes"); */
    /*     } */

    /*     // create staging buffer */
    /*     vk::Buffer stagingBuffer; */
    /*     MemoryInfo stagingBufferMemory; */
    /*     vk.createBuffer(indexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory); */

    /*     // fill staging buffer */
    /*     void* data; */
    /*     vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, indexBufferSize, NO_MEM_FLAGS, &data); */
    /*     if (result != vk::Result::eSuccess) { */
    /*         throw getVkException(result, "Failed to map staging buffer", "Renderer2D::fillIndexBufferWithShapes"); */
    /*     } */
    /*     uint32_t* idata = reinterpret_cast<uint32_t*>(data); */
    /*     size_t offset = 0; */
    /*     for (auto it = shapes.begin(); it != shapes.end(); it++) { */
    /*         rLog("fillIndexBufferWithShapes: copying index buffer nr", it - shapes.begin(), "-", it->getIndices(), "to address:", long(idata + offset), " offset:", offset); */
    /*         memcpy(idata+offset, it->getIndices().data(), it->getIndices().size() * sizeof(uint32_t)); */
    /*         offset += it->getIndices().size(); */
    /*     } */
    /*     rLog.log0("fillIndexBufferWithShapes: indices count:", shapesIndicesCount); */
    /*     vk.getDevice().unmapMemory(stagingBufferMemory.memory); */

    /*     // fill index buffer */
    /*     vk.copyBuffer(stagingBuffer, indexBuffer, indexBufferSize); */
    /*     vk.destroyBuffer(stagingBuffer, stagingBufferMemory); */
        
    /* } */


    void Renderer2D::drawShape(Shape* shape, uint32_t imageIndex, uint32_t currentFrame) {
        /* rLog.log0("drawShape:", shape->getVertexBufferInfo(), shape->getIndexBufferInfo()); */
        commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[R2Render].pipeline);

        uint32_t descriptorCount = 1;
        uint32_t firstSet = 0;
        uint32_t dynamicOffsetCount = 0;
        uint32_t* dynamicOffsets = nullptr;
        commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[R2Render].layout, firstSet, descriptorCount, &textureManager.getDescriptorSet(),  dynamicOffsetCount, dynamicOffsets);


        vk::Buffer vertexBuffers[] = { shapeBufferManager.getVertexBuffer(shape->getVertexBufferInfo()) };
        vk::DeviceSize vertexOffsets[] = { shape->getVertexBufferInfo().offset };
        uint32_t bindingCount = 1;
        uint32_t binding = 0;

        commandBuffers[currentFrame].bindVertexBuffers(binding, bindingCount, vertexBuffers, vertexOffsets);

        vk::Buffer indexBuffers = { shapeBufferManager.getIndexBuffer(shape->getIndexBufferInfo()) };
        vk::DeviceSize indexOffset = shape->getIndexBufferInfo().offset;
        // TODO use correct index type!
        commandBuffers[currentFrame].bindIndexBuffer(indexBuffers, indexOffset, vk::IndexType::eUint32);

        uint32_t pushConstant = shape->getTexureAtlasIndex();
        vk::ShaderStageFlags flags(vk::ShaderStageFlagBits::eFragment);
        commandBuffers[currentFrame].pushConstants(pipelines[R2Render].layout, flags, NO_OFFSET, sizeof(uint32_t), &pushConstant);

        uint32_t indexCount = shape->getIndexBufferInfo().count;
        uint32_t instanceCount = 1;
        uint32_t firstIndex = 0;
        uint32_t firstInstance = 0;
        commandBuffers[currentFrame].drawIndexed(indexCount, instanceCount, firstIndex, NO_OFFSET, firstInstance);
    }

    void Renderer2D::initializeShape(Shape* shape) {
        // make indices valid
        /* shape->setIndexOffset(shapesVerticesCount); */
        shape->normalizeVertices(vk.getScExtent().width, vk.getScExtent().height);
        shape->setTextureCoordinates(textureManager);
        shape->initalizeBufferInfos(shapeBufferManager);
    }

}  // namespace gz::vlk
