#include "renderer3D.hpp"

#include "vertex.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_instance.hpp"
#include "texture_manager.hpp"
#include "exceptions.hpp"
#include "vk_enum_string.h"
#include "vulkan_settings.hpp"

#include <gz-util/file_io.hpp>

#include <cstring>
#include <chrono>
#include <vulkan/vulkan_enums.hpp>
#include <vulkan/vulkan_structs.hpp>

constexpr uint32_t DS_MVP_BINDING = 0;

/* #define RENDERER_3D_RENDER_TO_SWAP_CHAIN */
#define RENDERER_3D_RENDER_TO_OWN_IMAGES
#define RENDERER_3D_COPY_TO_SWAP_CHAIN

// options sanity checks
#ifdef RENDERER_3D_RENDER_TO_SWAP_CHAIN
    #ifdef RENDERER_3D_RENDER_TO_OWN_IMAGES
        static_assert(false, "RENDERER_3D_RENDER_TO_SWAP_CHAIN and RENDERER_3D_RENDER_TO_OWN_IMAGES defined at the same time");
    #endif
    #ifdef RENDERER_3D_COPY_TO_SWAP_CHAIN
        static_assert(false, "RENDERER_3D_RENDER_TO_SWAP_CHAIN and RENDERER_3D_COPY_TO_SWAP_CHAIN defined at the same time");
    #endif
    #ifdef RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        static_assert(false, "RENDERER_3D_RENDER_TO_SWAP_CHAIN and RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES defined at the same time");
    #endif
#elif defined RENDERER_3D_RENDER_TO_OWN_IMAGES
/*     #ifndef RENDERER_3D_COPY_TO_SWAP_CHAIN */
/*         #define RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES */
/*     #endif */
#else
    static_assert(false, "Either RENDERER_3D_RENDER_TO_OWN_IMAGES or RENDERER_3D_RENDER_TO_SWAP_CHAIN must be defined");
#endif

namespace gz::vlk {
//
// INIT & CLEANUP
//
    Renderer3D::Renderer3D(VulkanInstance& instance, TextureManager& textureManager) : 
        Renderer(instance, textureManager) {
        LogCreateInfo logCI{
        	.logfile = "renderer3D.log",
        	.storeLog = false,
        	.prefix = "3D-Renderer",
        	.prefixColor = Color::LI_CYAN,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        rLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&Renderer3D::cleanup, this));
        vk.registerSwapChainRecreateCallback(std::bind(&Renderer3D::swapChainRecreateCallback, this));

        vk.createCommandBuffers(commandBuffers);
        const size_t vertexCount = 500000;
        const size_t indexCount = 500000;
        vk.createVertexBuffer<Vertex3D>(vertexCount, vertexBuffer, vertexBufferMemory, vertexBufferSize);
        vk.createIndexBuffer<uint32_t>(indexCount, indexBuffer, indexBufferMemory, indexBufferSize);

        // TODO
        loadModel();

        createUniformBuffers();
        createDescriptorResources();
        createDescriptorSetMvp();
        initSwapChainDependantResources();
        VulkanInstance::registerObjectUsingVulkan("Renderer3D",
                &vertexBuffer, &vertexBufferMemory.memory, 
                &indexBuffer, &indexBufferMemory.memory,
                &renderPass, &pipelines[R3Render].pipeline, //&pipelines[blendToSC].pipeline,
                &descriptorPool, 
                &mvpDSL, &sampleFromRenderImageDSL,
                &mvpDS, &sampleFromRenderImageDS,
                &imageSampler, &framebuffers, &renderImages, &renderImageViews,  // missing: &renderImageMemory, 
                &commandBuffers);
        rLog.log1("Created Renderer3D");
    }

    void Renderer3D::cleanup() {
        rLog.log0("cleanup: cleaning up");
        // UPDATE DOC ON CHANGES!
        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk.destroyBuffer(uniformBuffers[i], uniformBuffersMemory[i]);
        }
        vk.getDevice().destroyDescriptorSetLayout(mvpDSL, NO_ALLOC);
        vk.getDevice().destroyDescriptorPool(descriptorPool, NO_ALLOC);

#ifdef RENDERER_3D_SAMPLE_FROM_RENDER_IMAGES
        vk.getDevice().destroySampler(imageSampler);
#endif

        cleanupSwapChainDependantResources();
        cleanup_();
    }
//
// SWAPCHAIN DEPENDANT
//
    void Renderer3D::initSwapChainDependantResources() {
        // UPDATE DOC ON CHANGES!
        createRenderPass();
        createDepthImage();
        createImageResources();
#ifdef RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        createDescriptorSetSampler();
#endif
        vk.createFramebuffers(framebuffers, renderImageViews, renderPass, depthImageView);
        createPipelines();

    }


    void Renderer3D::cleanupSwapChainDependantResources() {
        // destroy pipelines
        pipelines.clear(vk.getDevice());

        vk.destroyFramebuffers(framebuffers);

        vk.destroyImageView(depthImageView);
        vk.destroyImage(depthImage, depthImageMemory);

#ifdef RENDERER_3D_RENDER_TO_OWN_IMAGES
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.destroyImageView(renderImageViews[i]);
            vk.destroyImage(renderImages[i], renderImageMemory[i]);
        }
#endif
        vk.getDevice().destroyRenderPass(renderPass, nullptr);
    }


    void Renderer3D::swapChainRecreateCallback() {
        cleanupSwapChainDependantResources();
        initSwapChainDependantResources();
    }


//
// IMAGES
//
    void Renderer3D::createImageResources() {
        renderImageViews.resize(vk.getScImages().size());
#ifdef RENDERER_3D_RENDER_TO_OWN_IMAGES
        renderImages.resize(vk.getScImages().size());
        renderImageMemory.resize(vk.getScImages().size());
        vk::ImageUsageFlags usage = vk::ImageUsageFlagBits::eColorAttachment;
    #ifdef RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        usage |= vk::ImageUsageFlagBits::eSampled;
    #elif defined RENDERER_3D_COPY_TO_SWAP_CHAIN
        usage |= vk::ImageUsageFlagBits::eTransferSrc;
    #endif
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImage(vk.getScExtent().width, vk.getScExtent().height, vk.getScImageFormat(), vk::ImageTiling::eOptimal, usage, vk::MemoryPropertyFlagBits::eDeviceLocal, renderImages[i], renderImageMemory[i]);
            vk.createImageView(vk.getScImageFormat(), renderImages[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
    #ifdef RENDERER_3D_SAMPLE_FROM_RENDER_IMAGES
        vk.createTextureSampler(imageSampler);
    #endif
        rLog.log0("createImageResources: created images, views and sampler");
#elif defined RENDERER_3D_RENDER_TO_SWAP_CHAIN
        for (size_t i = 0; i < renderImageViews.size(); i++) {
            vk.createImageView(vk.getScImageFormat(), vk.getScImages()[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        rLog.log0("createImageResources: created image views");
#endif
        assert(renderImageViews.size() == vk.getScImages().size());
    }


//
// DEPTH
//
    void Renderer3D::createDepthImage() {
        vk::MemoryPropertyFlags memoryProperties = vk::MemoryPropertyFlagBits::eDeviceLocal;
        vk::ImageUsageFlags imageUsage = vk::ImageUsageFlagBits::eDepthStencilAttachment;
        vk::ImageTiling tiling = vk::ImageTiling::eOptimal;
        vk::Format depthFormat = vk.getDepthFormat();

        vk.createImage(vk.getScExtent().width, vk.getScExtent().height, depthFormat, tiling, imageUsage, memoryProperties, depthImage, depthImageMemory);
        vk.createImageView(depthFormat, depthImage, depthImageView, vk::ImageAspectFlagBits::eDepth);

        vk.transitionImageLayout(depthImage, depthFormat, vk::ImageLayout::eUndefined, vk::ImageLayout::eDepthStencilAttachmentOptimal);
    }

//
// RENDER PASS
//
    void Renderer3D::createRenderPass() {
        vk::AttachmentDescription2 colorBlendAD {
            .format = vk.getScImageFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eStore,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
#ifdef RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
            .finalLayout = vk::ImageLayout::eShaderReadOnlyOptimal,
#elif defined RENDERER_3D_COPY_TO_SWAP_CHAIN
            .finalLayout = vk::ImageLayout::eTransferSrcOptimal,
#else
            .finalLayout = vk::ImageLayout::eColorAttachmentOptimal,
#endif
        };

        vk::AttachmentReference2 colorBlendAR {
            .attachment = 0,
            .layout = vk::ImageLayout::eColorAttachmentOptimal,
        };

        vk::AttachmentDescription2 depthAD {
            .format = vk.getDepthFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eDontCare,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
            .finalLayout = vk::ImageLayout::eDepthStencilAttachmentOptimal,
        };

        vk::AttachmentReference2 depthAR {
            .attachment = 1,
            .layout = vk::ImageLayout::eDepthStencilAttachmentOptimal,
        };

        vk::SubpassDescription2 subpass {
            .pipelineBindPoint = vk::PipelineBindPoint::eGraphics,
            .colorAttachmentCount = 1,
            .pColorAttachments = &colorBlendAR,
            .pDepthStencilAttachment = &depthAR,
        };

        // wait until image is available
        vk::SubpassDependency2 colorAttachmentSD {
            .srcSubpass = VK_SUBPASS_EXTERNAL,
            .dstSubpass = 0,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests,
            .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests,
            .srcAccessMask = NO_ACC_FLAGS,
            .dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentWrite,
        };

#ifdef RENDERER_3D_COPY_TO_SWAP_CHAIN
        // image layout transition to transfer dst
        vk::SubpassDependency2 layoutTransitionSD {
            .srcSubpass = 0,
            .dstSubpass = VK_SUBPASS_EXTERNAL,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eTransfer,
            .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
            .dstAccessMask = vk::AccessFlagBits::eTransferRead,
            .dependencyFlags = vk::DependencyFlagBits::eByRegion,
        };
#endif

        /* vk::SubpassDependency2 dependency { */
        /*     .srcSubpass = VK_SUBPASS_EXTERNAL, */
        /*     .dstSubpass = 0, */
        /*     .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests, */
        /*     .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests, */ 
        /*     .srcAccessMask = NO_ACC_FLAGS, */
        /*     .dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentWrite, */
        /* }; */

        std::vector<vk::AttachmentDescription2> attachments = { colorBlendAD, depthAD, };
        std::vector<vk::SubpassDependency2> dependencies = { colorAttachmentSD
#ifdef RENDERER_3D_COPY_TO_SWAP_CHAIN
            , layoutTransitionSD
#endif
        };
        vk::RenderPassCreateInfo2 renderPassCI;
        /* renderPassCI.setCorrelatedViewMasks() */
        renderPassCI.setAttachments(attachments);
        renderPassCI.setDependencies(dependencies);
        renderPassCI.setSubpasses(subpass);
        vk::Result result = vk.getDevice().createRenderPass2(&renderPassCI, nullptr, &renderPass); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Could not create render pass", "Renderer3D::createRenderPass");
        }
        rLog.log0("createRenderPass: Created render pass.");
    }


	void Renderer3D::createPipelines() {
        // VERTEX 3D PIPELINE
        std::vector<vk::DescriptorSetLayout> descriptorSetLayouts = { textureManager.getDescriptorSetLayout(), mvpDSL };
        /* std::vector<vk::DescriptorSetLayout> descriptorSetLayouts = { mvpDSL }; */

        // specialization constant for sampler2D array length = atlas count
        vk::SpecializationMapEntry specME {
            .constantID = 0,
            .offset = 0,
            .size = sizeof(uint32_t),
        };
        uint32_t atlasCount = textureManager.getAtlasCount();
        vk::SpecializationInfo specI;
        specI.setData<uint32_t>(atlasCount);
        specI.setMapEntries(specME);

        auto vertShaderCode = readBinaryFile("shaders/shader.vert.spv");
        auto fragShaderCode = readBinaryFile("shaders/shader.frag.spv");

        vk::ShaderModule vertShaderModule = vk.createShaderModule(vertShaderCode);
        vk::ShaderModule fragShaderModule = vk.createShaderModule(fragShaderCode);

        std::vector<vk::PipelineShaderStageCreateInfo> shaderStages;

        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eVertex,
        	.module = vertShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = nullptr,
        });
        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eFragment,
        	.module = fragShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = &specI,
        });

        // custom rasterizationState with clockwise front face
        vk::PipelineRasterizationStateCreateInfo rasterizationStateCI {
        	.depthClampEnable = VK_FALSE,
        	.rasterizerDiscardEnable = VK_FALSE,
        	.polygonMode = vk::PolygonMode::eFill,
        	.cullMode = vk::CullModeFlagBits::eBack,
        	.frontFace = vk::FrontFace::eClockwise,
        	.depthBiasEnable = VK_FALSE,
        	.lineWidth = 1.0f,
            /* .depthBiasConstantFactor = 0.0f, */
            /* .depthBiasClamp = 0.0f, */
            /* .depthBiasSlopeFactor = 0.0f, */
        };
        vk::GraphicsPipelineCreateInfo pipelineCI {
            .pRasterizationState = &rasterizationStateCI,
            .renderPass = renderPass,
            .subpass = 0,
        };
        pipelineCI.setStages(shaderStages);

        std::vector<vk::PushConstantRange> pushConstantRanges;
        pushConstantRanges.emplace_back(vk::PushConstantRange{
            .stageFlags = vk::ShaderStageFlagBits::eFragment,
            .size = sizeof(uint32_t),
        });
        vk.createGraphicsPipeline<Vertex3D>(std::move(pipelineCI), descriptorSetLayouts, pushConstantRanges, true, pipelines[R3Render]);

        vk.getDevice().destroyShaderModule(vertShaderModule, nullptr);
        vk.getDevice().destroyShaderModule(fragShaderModule, nullptr);


        // BLIT TO SWAP CHAIN PIPELINE
        /* auto fragShaderCode2 = readBinaryFile("shaders/blend.frag.spv"); */
        /* vk::ShaderModule fragShaderModule2 = vk.createShaderModule(fragShaderCode); */

        /* std::vector<vk::PipelineShaderStageCreateInfo> shaderStages2; */
        /* shaderStages2.emplace_back(vk::PipelineShaderStageCreateInfo { */
        /* 	.stage = vk::ShaderStageFlagBits::eFragment, */
        /* 	.module = fragShaderModule2, */
        /* 	.pName = "main", */
        /* 	.pSpecializationInfo = nullptr, */
        /* }); */

        /* vk::GraphicsPipelineCreateInfo pipelineCI2 { */
        /*     .renderPass = renderPass, */
        /*     .subpass = 1, */
        /* }; */
        /* pipelineCI2.setStages(shaderStages2); */

        /* vk.createGraphicsPipeline<Vertex3D>(std::move(pipelineCI2), nullptr, false, pipelines[blendToSC]); */

        /* vk.getDevice().destroyShaderModule(fragShaderModule2, nullptr); */
		
	}



//
// DESCRIPTORS
//
    void Renderer3D::createDescriptorResources() {
        // LAYOUTS
        std::vector<vk::DescriptorSetLayoutBinding> mvpDSLB;
        // SET 1: uniform buffer object
        mvpDSLB.emplace_back(vk::DescriptorSetLayoutBinding {
            .binding = DS_MVP_BINDING,
            .descriptorType = vk::DescriptorType::eUniformBuffer,
            .descriptorCount = 1,
            .stageFlags = vk::ShaderStageFlagBits::eVertex,
        });

        vk.createDescriptorSetLayout(mvpDSLB, mvpDSL);

#ifdef RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
        createDescriptorSetLayoutSampler(0);
#endif
        // POOL
        std::vector<vk::DescriptorPoolSize> poolSizes;
        poolSizes = { 
            { .type = vk::DescriptorType::eUniformBuffer, .descriptorCount = vk.getMaxFramesInFlight() },
#ifdef RENDERER_3D_ENABLE_SAMPLE_FROM_RENDER_IMAGES
            { .type = vk::DescriptorType::eCombinedImageSampler, .descriptorCount = static_cast<uint32_t>(vk.getScImages().size()) },
#endif
        };
        uint32_t maxSets = 0;
        for (auto& size : poolSizes) {
            maxSets += size.descriptorCount;
        }
        vk.createDescriptorPool(poolSizes, maxSets, descriptorPool);

        rLog.log0("createDescriptorResources: Created descriptor layouts and pool");
    }


	void Renderer3D::createDescriptorSetMvp() {
        std::vector<vk::DescriptorSetLayout> layouts1(vk.getMaxFramesInFlight(), mvpDSL);
        vk.createDescriptorSets(layouts1, descriptorPool, mvpDS);

        assert(uniformBuffers.size() == vk.getMaxFramesInFlight());
        assert(mvpDS.size() == vk.getMaxFramesInFlight());
        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk::DescriptorBufferInfo bufferI {
            	.buffer = uniformBuffers[i],
            	.offset = 0,
            	.range = VK_WHOLE_SIZE,  // sizeof(ModelViewProjection),
            };

            vk::WriteDescriptorSet descriptorW{
                .dstSet = mvpDS[i],
                .dstBinding = DS_MVP_BINDING,
                .dstArrayElement = 0,
                .descriptorCount = 1,
                .descriptorType = vk::DescriptorType::eUniformBuffer,
                .pBufferInfo = &bufferI,
            };
            /* descriptorW[0].pTexelBufferView = nullptr; */
            
            // write 1, copy 0
            vk.getDevice().updateDescriptorSets(descriptorW, nullptr);
        }  // for
        rLog.log0("createDescriptorSetMvp: Created", mvpDS.size(), "descriptor sets for model view projection");
    }


//
// MODEL
//
	void Renderer3D::loadModel() {
        // load model into VerticesAndIndices struct
        rLog.log1("Renderer3D: loadModel: loading model");
        /* vk.loadModel(settings::MODEL_PATH, model); */
        // TODO use correct type
        vk::DeviceSize requiredVertexBufferSize = model.vertices.size() * sizeof(Vertex3D);
        vk::DeviceSize requiredIndexBufferSize = model.indices.size() * sizeof(uint32_t);

        if (requiredVertexBufferSize > vertexBufferSize) { throw VkException("Renderer3D::loadModel: vertex buffer too small, required size = " + gz::toString(requiredVertexBufferSize)); }
        if (requiredIndexBufferSize > indexBufferSize) { throw VkException("Renderer3D::loadModel: index buffer too small = " + gz::toString(requiredIndexBufferSize)); }

        rLog.log0("Renderer3D: loadModel: filling vertex buffer");
        // copy to vertexBuffer
        // create staging buffer
        vk::Buffer stagingBuffer;
        MemoryInfo stagingBufferMemory;
        vk.createBuffer(requiredVertexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);
        // fill staging buffer
        void* data;
        vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, requiredVertexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer for vertices", "Renderer3D::loadModel");
        }

        memcpy(data, model.vertices.data(), requiredVertexBufferSize);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill vertex buffer 
        vk.copyBuffer(stagingBuffer, vertexBuffer, requiredVertexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);

        rLog.log0("Renderer3D: loadModel: filling index buffer");
        data = nullptr;
        // copy to index buffer
        vk.createBuffer(requiredIndexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);
        // fill staging buffer
        result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, requiredIndexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer for indices", "Renderer3D::loadModel");
        }
        memcpy(data, model.indices.data(), requiredIndexBufferSize);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill index buffer 
        vk.copyBuffer(stagingBuffer, indexBuffer, requiredIndexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);
	}

//
// RENDERING
//
    void Renderer3D::recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame) {
        vk::CommandBufferBeginInfo commandBufferBI;
        vk::Result result = commandBuffers[currentFrame].begin(commandBufferBI); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to begin 3D command buffer", "Renderer3D::recordCommandBuffer");
        }

        // clear
        std::array<vk::ClearValue, 2> clearValues{};
        vk::ClearColorValue clearColor{};
        clearColor.setFloat32({0.1f, 0.0f, 0.0f, 0.5f});
        clearValues[0].setColor(std::move(clearColor));
        clearValues[1].setDepthStencil({1.0f, 0});

        vk::RenderPassBeginInfo renderPassBI {
        	.renderPass = renderPass,
        	.framebuffer = framebuffers[imageIndex],
            .renderArea { .offset = { 0, 0 }, .extent = vk.getScExtent() }
        };
        renderPassBI.setClearValues(clearValues);

        commandBuffers[currentFrame].beginRenderPass(renderPassBI, vk::SubpassContents::eInline);

        commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[R3Render].pipeline);
        vk::Buffer vertexBuffers[] = { vertexBuffer };
        vk::DeviceSize offsets[] = {0};
        uint32_t bindingCount = 1;
        uint32_t firstBinding = 0;
        commandBuffers[currentFrame].bindVertexBuffers(firstBinding, bindingCount, vertexBuffers, offsets);
        // TODO use correct index type!
        commandBuffers[currentFrame].bindIndexBuffer(indexBuffer, NO_OFFSET, vk::IndexType::eUint32);

        uint32_t firstSet = 0;
        std::vector<vk::DescriptorSet> descriptorSets = { textureManager.getDescriptorSet(), mvpDS[currentFrame], };
        std::vector<uint32_t> dynamicOffsets;
        commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[R3Render].layout, firstSet, descriptorSets, dynamicOffsets);
        /* commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[R3Render].layout, firstSet, mvpDS[currentFrame], dynamicOffsets); */
        /* commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[R3Render].layout, firstSet, descriptorSets, dynamicOffsets); */

        int instanceCount = 1;
        int firstIndex = 0;
        int firstInstance = 0;
        commandBuffers[currentFrame].drawIndexed(static_cast<uint32_t>(model.indices.size()), instanceCount, firstIndex, NO_OFFSET, firstInstance);

        /* commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[blendToSC].pipeline); */
        /* commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[blendToSC].layout, firstSet, sampleFromRenderImageDS[currentFrame], {}); */
        /* // draw hardcoded fullscreen quad */
        /* commandBuffers[currentFrame].draw(6, 1, 0, 0); */
        commandBuffers[currentFrame].endRenderPass();

#ifdef RENDERER_3D_COPY_TO_SWAP_CHAIN
        vk.copyImageToImage(commandBuffers[currentFrame], renderImages[imageIndex], vk.getScImages()[imageIndex], vk.getScExtent());
#endif
        result = commandBuffers[currentFrame].end();
        if (result != vk::Result::eSuccess) {
            rLog.error("Failed to record 3D - command buffer", "VkResult:", result);
            throw getVkException(result, "Failed to record 3D - command buffer", "Renderer3D::recordCommandBuffer");
        }
        vk.submitThisFrame(commandBuffers[currentFrame]);
    }
//
// RENDERING
//
    void Renderer3D::drawFrame(uint32_t imageIndex) {
        commandBuffers[vk.getCurrentFrame()].reset(NO_CMD_RESET_FLAGS);
        /* recordCommandBuffer(imageIndex, vk.getCurrentFrame()); */
        updateUniformBuffer();
        recordCommandBuffer(imageIndex, vk.getCurrentFrame());
    }


    void Renderer3D::createUniformBuffers() {
        vk::DeviceSize bufferSize = sizeof(ModelViewProjection);

        uniformBuffers.resize(vk.getMaxFramesInFlight());
        uniformBuffersMemory.resize(vk.getMaxFramesInFlight());

        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk.createBuffer(bufferSize, vk::BufferUsageFlagBits::eUniformBuffer, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, uniformBuffers[i], uniformBuffersMemory[i]);
        }
    }


    void Renderer3D::updateUniformBuffer() {
        static auto startTime = std::chrono::high_resolution_clock::now();
        auto currentTime = std::chrono::high_resolution_clock::now();
        float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();

        // TODO use push constant instead of ubo
        ModelViewProjection mvp{};
        mvp.model = glm::rotate(glm::mat4(1.0f), time * std::numbers::pi_v<float> / 2, glm::vec3(0.0f, 0.0f, 1.0f));
        mvp.view = glm::lookAt(glm::vec3(2.0f, 2.0f, 2.0f), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f));
        mvp.projection = glm::perspective(glm::radians(45.0f), static_cast<float>(vk.getScExtent().width) / vk.getScExtent().height, 1.0f, 10.0f);
        /* ubo.model = glm::mat4(1); */
        /* ubo.view = glm::mat4(1); */
        /* ubo.projection = glm::mat4(1); */
        /* ubo.projection[1][1] *= -1;  // y coordinate inverted in opengl */
        MemoryInfo& uniformBufferMI = uniformBuffersMemory[vk.getCurrentFrame()];
        void* data;
        vk::Result result = vk.getDevice().mapMemory(uniformBufferMI.memory, uniformBufferMI.offset, sizeof(mvp), NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer", "Renderer3D::updateUniformBuffer");
        }
        memcpy(data, &mvp, sizeof(mvp));
        vk.getDevice().unmapMemory(uniformBufferMI.memory);
    }

}  // namespace gz::vlk
