#pragma once

#include "buffer_manager.hpp"
#include "vertex.hpp"
#include <cstdint>

namespace gz::vlk {
    // defined in texture_manager.hpp
    class TextureManager;
    /**
     * @brief Base class for shapes
     * @details
     *  In this implementaiton, a shape is a 2D object made of several vertices.
     *  Each shape has its own vertex and index buffers, which are managed by a BufferManager.
     *  Their texture is managed by a TextureManager.
     * @todo free resources and rule of 5
     */
    class Shape {
        public:
			const std::vector<Vertex2D>& getVertices() const { return vertices; }
			const std::vector<uint32_t>& getIndices() const { return indices; }
			const std::string& getTexture() const { return texture; }
			uint32_t getTexureAtlasIndex() const { return textureAtlasIndex; }
			const BufferInfo& getVertexBufferInfo() const { return vertexBufferInfo; }
			const BufferInfo& getIndexBufferInfo() const { return indexBufferInfo; }

            /**
             * @brief Add an offset to all indices (useful when putting multiple shapes in the same vertex buffer)
             */
            void setIndexOffset(uint32_t offset);
            /**
             * @brief Normalize the vertices, so that (1, 1) is (width, height)
             */
            void normalizeVertices(float width, float height);
            virtual void setTextureCoordinates(glm::vec2 topLeft, glm::vec2 bottomRight) {};
            virtual void setTextureCoordinates(TextureManager& textureManager) {};
            bool buffersInitalized() const { return vertexBufferInfo.index != BUFFER_NOT_INITIALIZED; }
            void initalizeBufferInfos(BufferManager<Vertex2D, uint32_t>& bufferManager);
            virtual ~Shape() {};
        protected:
            std::string texture = "texture.png";
            uint32_t textureAtlasIndex;
            std::vector<Vertex2D> vertices;
            std::vector<uint32_t> indices;

            BufferInfo vertexBufferInfo;
            BufferInfo indexBufferInfo;
    };

    class Rectangle : public Shape {
        public:
            Rectangle(float top, float left, uint32_t width, uint32_t height, glm::vec3 color, std::string&& texture);
            void setTextureCoordinates(glm::vec2 topLeft, glm::vec2 bottomRight) override;
            /**
             * @brief Get the correct texture coordinates from a TextureManager
             * @details
             *  If the texture is "atlas", the texture coordinates will remain at (0,0), (0,1), (1, 1), (1, 0) and
             *  thus show the entire atlas.
             */
            void setTextureCoordinates(TextureManager& textureManager) override;
        private:
            float top, left;
            uint32_t width, height;
            glm::vec3 color;
            void generateVertices();
    };
}  // namespace gz::vlk
