#include "texture_atlas.hpp"

#include "vulkan_instance.hpp"
#include "exceptions.hpp"

#include "stb_image.h"

#include <cstdint>
#include <cstdio>
#include <cstring>
#include <gz-util/exceptions.hpp>
#include <gz-util/util/string_concepts.hpp>
#include <gz-util/util/string_conversion.hpp>
#include <ratio>

#include <glm/glm.hpp>

namespace gz::vlk {
std::string TextureImageArea::toString() const {
    return "<(" + gz::toString(x) + "," + gz::toString(y) + "), (" + gz::toString(width) + "x" + gz::toString(height) + ")>";
}

//
// INIT & CLEANUP
//
TextureAtlas::TextureAtlas(VulkanInstance& instance, uint16_t slotWidth, uint16_t slotHeight, uint16_t slotCountX, uint16_t slotCountY)
    : vk(instance), 
        slotWidth(slotWidth), slotHeight(slotHeight), 
        slotCountX(slotCountX), slotCountY(slotCountY)
{
#ifdef LOG_LEVEL_0
    LogCreateInfo logCI{
    	.logfile = "textureAtlas.log",
    	.storeLog = false,
    	.prefix = "TextureAtlas (" + gz::toString(slotCountX) + "x" + gz::toString(slotCountY) + ")X(" + gz::toString(slotWidth) + "x" + gz::toString(slotHeight) + ")",
    	.prefixColor = Color::GREEN,
    	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
    };
    tLog = Log(std::move(logCI));
#endif

    // whole image
    freeAreas.insert({ 0, 0, slotCountX, slotCountY});
    createImageResources();
    /* vk.registerCleanupCallback(std::bind(&TextureAtlas::cleanup, this)); */
    VulkanInstance::registerObjectUsingVulkan(std::string("TextureAtlas-") + gz::toString(slotWidth) + "x" + gz::toString(slotHeight),
            &textureImage, &textureImageMemory.memory, &textureImageView, &textureSampler);
}


void TextureAtlas::cleanup() {
#ifdef LOG_LEVEL_0
    tLog.log0("cleanup: cleaning up");
#endif
    vk.destroyTextureSampler(textureSampler);
    vk.destroyImageView(textureImageView);
    vk.destroyImage(textureImage, textureImageMemory);
}


void TextureAtlas::createImageResources() {
    vk.createImage(slotWidth * slotCountX, slotHeight * slotCountY, vk::Format::eR8G8B8A8Srgb, vk::ImageTiling::eLinear, 
            vk::ImageUsageFlagBits::eTransferDst | vk::ImageUsageFlagBits::eSampled, 
            vk::MemoryPropertyFlagBits::eDeviceLocal, 
            textureImage, textureImageMemory);
    vk::CommandBuffer cmdBuffer = vk.beginSingleTimeCommands(POOL_GRAPHICS);
    vk.transitionImageLayout(textureImage, vk::Format::eR8G8B8A8Srgb, vk::ImageLayout::eUndefined, vk::ImageLayout::eTransferDstOptimal, &cmdBuffer);
    vk::ImageSubresourceRange subresourceRange {
        .aspectMask = vk::ImageAspectFlagBits::eColor,
        .baseMipLevel = 0,
        .levelCount = 1,
        .baseArrayLayer = 0,
        .layerCount = 1,
    };
    cmdBuffer.clearColorImage(textureImage, vk::ImageLayout::eTransferDstOptimal, &settings::missingTextureColor, 1, &subresourceRange);
    vk.transitionImageLayout(textureImage, vk::Format::eR8G8B8A8Srgb, vk::ImageLayout::eTransferDstOptimal, vk::ImageLayout::eShaderReadOnlyOptimal, &cmdBuffer);
    vk.endSingleTimeCommands(cmdBuffer, POOL_GRAPHICS);

    vk.createImageView(vk::Format::eR8G8B8A8Srgb, textureImage, textureImageView, vk::ImageAspectFlagBits::eColor);
    vk.createTextureSampler(textureSampler);
}


//
// ADD TEXTURE
//
std::pair<glm::vec2, glm::vec2> TextureAtlas::addTexture(uint8_t* pixels, uint16_t width, uint16_t height) {
    if (freeAreas.empty()) {
        throw Exception("No texture slots available", "TextureAtlas::addTexture");
    }
    // needed slot count in x/y direction
    uint16_t slotsX, slotsY;
    // width / slotHeight (+ 1 if another slot is not fully filled)
    slotsX = width / slotWidth + ((width % slotWidth != 0) ? 1 : 0);
    slotsY = height / slotHeight + ((height % slotHeight != 0) ? 1 : 0);

    // find smallest area that fits the texture
    auto freeArea = freeAreas.end();
    for (auto it = freeAreas.begin(); it != freeAreas.end(); it++) {
        // if fits the texture
        if (it->width >= slotsX and it->height >= slotsY) {
            freeArea = it;
            break;
        }
    }
    // check if match
    if (freeArea == freeAreas.end()) {
        throw Exception("Not enough texture slots available to fit the texture", "TextureAtlas::addTexture");
    }

    // consume free area
    auto node = freeAreas.extract(freeArea);
    TextureImageArea& selectedArea = node.value();

    // topleft slot for texture
    uint16_t x = selectedArea.x;
    uint16_t y = selectedArea.y;
    // update selectedArea 
    uint16_t selectedAreaNewY = selectedArea.y + slotsY;
    uint16_t selectedAreaNewWidth = slotsX;
    uint16_t selectedAreaNewHeight = selectedArea.height - slotsY;
    // if split needed
    if (slotsX < selectedArea.width and slotsY < selectedArea.height) {
        // split so that the size difference between the new free area and the remaining free area is maximal
        TextureImageArea newFreeArea;
        newFreeArea.x = selectedArea.x + slotsX;  // right of new texture
        newFreeArea.y = selectedArea.y;
        newFreeArea.width = selectedArea.width - slotsX;
        // try full height first first
        newFreeArea.height = selectedArea.height;

        // size difference with (newArea.height == full height) < size difference with (newArea.height == textureHeight)
        int sizeDiffNewAreaFullHeight = static_cast<int>(newFreeArea.width * newFreeArea.height) - selectedAreaNewWidth * selectedAreaNewHeight;
        int sizeDiffNewAreaTextureHeight = static_cast<int>(newFreeArea.width * (selectedArea.height - slotsY)) - selectedArea.width * selectedAreaNewHeight;
        if (sizeDiffNewAreaFullHeight * sizeDiffNewAreaFullHeight < sizeDiffNewAreaTextureHeight * sizeDiffNewAreaTextureHeight) {
            newFreeArea.height = slotsY;
            selectedAreaNewWidth = selectedArea.width;
        }
        // insert new area
        freeAreas.insert(std::move(newFreeArea));
    }

    if (selectedAreaNewHeight > 0 and selectedAreaNewWidth > 0) {
        // insert selected area with updated width, height and y
        selectedArea.width = selectedAreaNewWidth;
        selectedArea.height = selectedAreaNewHeight;
        selectedArea.y = selectedAreaNewY;
        freeAreas.insert(std::move(node));
    }
    mergeFreeAreas();
#ifdef LOG_LEVEL_0
    tLog.log0("addTexture: Adding texture at position x,y=(", x, y, "), with slotCountX,Y=(", slotsX, slotsY, "), with dimensions w,h=(", width, height, ")");
#endif
    blitTextureOnImage(pixels, x, y, width, height);
    // topleft normalized: slot / slotCount
    // bottomright normalized: (slot + (textureSize/slotCountize)) / slotCount
    return { glm::vec2(static_cast<float>(x) / slotCountX, static_cast<float>(y) / slotCountY), 
             glm::vec2((static_cast<float>(x) + static_cast<float>(width) / slotWidth) / slotCountX,
                (static_cast<float>(y) + static_cast<float>(height) / slotHeight) / slotCountY) };
}


void TextureAtlas::blitTextureOnImage(uint8_t* pixels, uint16_t slotX, uint16_t slotY, uint16_t textureWidth, uint16_t textureHeight) {
    constexpr size_t BYTES_PER_PIXEL = 4;
    vk::DeviceSize imageSize = textureWidth * textureHeight * BYTES_PER_PIXEL;
    vk::Buffer stagingBuffer;
    MemoryInfo stagingBufferMemory;
    vk.createBuffer(imageSize, vk::BufferUsageFlagBits::eTransferSrc, 
            vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, 
            stagingBuffer, stagingBufferMemory);
    void* data;
    vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, imageSize, vk::MemoryMapFlags(), &data);
    if (result != vk::Result::eSuccess) {
        throw getVkException(result, "Failed to map staging buffer", "TextureAtlas::blitTextureOnImage");
    }
    
    memcpy(data, pixels, static_cast<size_t>(imageSize));
    vk.getDevice().unmapMemory(stagingBufferMemory.memory);

    vk.transitionImageLayout(textureImage, vk::Format::eR8G8B8A8Srgb, vk::ImageLayout::eShaderReadOnlyOptimal, vk::ImageLayout::eTransferDstOptimal);
    vk.copyBufferToImage(stagingBuffer, textureImage, 
            static_cast<int32_t>(slotWidth * slotX), static_cast<int32_t>(slotHeight * slotY),  // offset
            static_cast<uint32_t>(textureWidth), static_cast<uint32_t>(textureHeight));  // dimensions
    vk.transitionImageLayout(textureImage, vk::Format::eR8G8B8A8Srgb, vk::ImageLayout::eTransferDstOptimal, vk::ImageLayout::eShaderReadOnlyOptimal);
    
    vk.destroyBuffer(stagingBuffer, stagingBufferMemory);
}


void TextureAtlas::mergeFreeAreas() {
begin:
#ifdef LOG_LEVEL_0
    tLog.log0("mergeFreeAreas before merge:", freeAreas);
#endif
    // surely not the most efficient way, but it should only be run at load time and thus not be too problematic
    // iterate over the areas for each area
    for (auto it = freeAreas.begin(); it != freeAreas.end(); it++) {
        for (auto it2 = freeAreas.begin(); it2 != freeAreas.end(); it2++) {
            if (it == it2) { continue; }
            // if same y and same height and next to each other
            if (it->y == it2->y and it->height == it2->height and it2->x == it->x + it->width) {
                auto area1 = freeAreas.extract(it);  // node
                area1.value().width += it2->width;
                freeAreas.insert(std::move(area1));
                freeAreas.erase(it2);
                goto begin;  // start over since the iterators are invalid and new merges might be possible
            }
            // if same x and same with and below/above each other
            if (it->x == it2->x and it->width == it2->width and it2->y == it->y + it->height) {
                auto area1 = freeAreas.extract(it);  // node
                area1.value().height += it2->height;
                freeAreas.insert(std::move(area1));
                freeAreas.erase(it2);
                goto begin;  // start over since the iterators are invalid and new merges might be possible
            }
        }
    }
#ifdef LOG_LEVEL_0
    tLog.log0("mergeFreeAreas after merge:", freeAreas);
#endif
}

std::string TextureAtlas::toString() const {
    return "TextureAtlas(SlotSize: " + gz::toString(slotWidth) + "x" + gz::toString(slotHeight)
        + ", SlotCount: " + gz::toString(slotCountX) + "x" + gz::toString(slotCountY) + ", FreeAreas: " + gz::toString(freeAreas) + ")";
}


}  // namespace gz 
