#pragma once

#include "vertex.hpp"

#include <chrono>
#include <optional>
#include <set>
#include <type_traits>
#include <unordered_map>

#define VULKAN_HPP_NO_CONSTRUCTORS
#define VULKAN_HPP_NO_EXCEPTIONS
#include <vulkan/vulkan.hpp>

#ifdef GZ_UTIL_STRING_CONCEPTS
    static_assert(false, "gz-util/util/string_conversion.hpp is included before vulkan_util.hpp!");
#endif

template<typename T>
concept VkToString = requires(const T& t) {
    { vk::to_string(t) };
};

template<VkToString T>
inline std::string toString(const T& t) {
    return vk::to_string(t);
}

namespace gz::vlk {
    /**
     * @name Convenience structs
     * @details
     *  Structures and functionas making usiing the vulkan api more convenient
     * @{
     */
    /**
     * @brief Struct containing all physical device features structs
     * @details
     *  Make sure to call vkGetPhysicalDeviceFeatures2 with the f member, not this struct itself
     * @todo Redo with vulkan.hpp and vk::StructureChain
     */
    struct PhysicalDeviceFeatures {
        /**
         * @brief Initialize the structs with sType and pNext chain f -> f11 -> f12 -> f13
         */
        PhysicalDeviceFeatures();
        VkPhysicalDeviceFeatures2 f {};
        VkPhysicalDeviceVulkan11Features f11{};
        VkPhysicalDeviceVulkan12Features f12{};
        VkPhysicalDeviceVulkan13Features f13{};
        /**
         * @brief Check if all features enabled in requiredFeatures are enabled in this
         */
        bool requiredFeaturesAvailable(const PhysicalDeviceFeatures& requiredFeatures) const;
    };

    struct Pipeline {
        vk::Pipeline pipeline;
        vk::PipelineLayout layout;
        std::vector<vk::ShaderModule> shaderModules;
        void operator=(const Pipeline& other) = delete;
    };
    template<typename T>
    concept PipelineKeyType = requires { std::unordered_map<T, Pipeline>(); };

    /**
     * @brief Map holding pipelines
     * @details
     *  A map holding Pipeline objects.
     *  I suggest declaring an enum for all pipelines and using this enum as key type.
     *  This way you can use a pipeline "by its name" without having to make it a member variable
     *
     *  @section Usage
     *   Using operator[] creates a default initialized Pipeline struct.
     *   First, create the shader
     *   Pass this to
     *   @code
     *    enum MyPipelines { pl1, pl2, ... };
     *    MyClass {
     *        ...
     *        PipelineContainer<MyPipelines> pipelines;
     *        ...
     *    };
     *
     *    MyClass::createPipeline() {
     *        // operator[] default initalizes a Pipeline struct
     *        // see text below for explanation
     *        if (pipelines[pl1].shaderModules.size() == 0) {
     *            // will create shader modules in pipelines[pl1].shaderModules
     *            // TODO
     *            VulkanInstance::createShaderModules({"shader.vert.spv",  }, ..., pipelines[pl1]);
     *        }
     *        // create pipeline and layout in pipelines[pl1].layout, pipelines[pl1].pipeline
     *        VulkanInstance::createGraphicsPipeline(..., pipelines[pl1])
     *    }
     *
     *    MyClass::doStuff() {
     *        ...
     *        cmdBuffer.bindPipeline(pipelines[pl1].pipeline);
     *        ...
     *    }
     *
     *    MyClass::swapChainRecreateCallback() {
     *        device.destroyPipeline(pipelines[pl1].pipeline);
     *        device.destroyPipelineLayout(pipelines[pl1].layout);
     *        createPipeline();
     *    }
     *
     *    MyClass::cleanup() {
     *        pipelines.clear(device);
     *    }
     *
     *
     *
     *   @endcode
     *
     *   The reason why the shader modules are stored together with the pipelines 
     *   (for now, maybe this will change when I add pipeline caching)
     *   is that it can improve pipeline recreation times (eg. when window size changes).
     *   If the shader modules stay valid (meaning that the specialization values stay the same)
     *   you can just reuse them. 
     *
     */
    template<PipelineKeyType T> 
    class PipelineContainer {
        public:
            using iterator = typename std::unordered_map<T, Pipeline>::iterator;
            Pipeline& operator[](const T& key) { return pipelines[key]; }
            /**
             * @brief Destroy the pipeline, layout and shader modules and then remove the handles from the underlying map
             */
            void erase(const T& key, const vk::Device& device, const vk::AllocationCallbacks* pAllocator=nullptr);
            /**
             * @brief Destroy the pipeline, layout and shader modules and then remove the handles from the underlying map
             */
            iterator erase(const iterator& it, const vk::Device& device, const vk::AllocationCallbacks* pAllocator=nullptr);
            /**
             * @brief Destroy all pipelines, layouts and shader modules and then remove the handles from the underlying map
             */
            void clear(const vk::Device& device, const vk::AllocationCallbacks* pAllocator=nullptr);
            iterator begin() { return pipelines.begin(); }
            iterator end() { return pipelines.end(); }
            size_t size() const { return pipelines.size(); }
        private:
            std::unordered_map<T, Pipeline> pipelines;
    };

    template<PipelineKeyType T>
	void PipelineContainer<T>::erase(const T& key, const vk::Device& device, const vk::AllocationCallbacks* pAllocator) {
        device.destroyPipeline(pipelines[key].pipeline, pAllocator);
        device.destroyPipelineLayout(pipelines[key].layout, pAllocator);
		pipelines.erase(pipelines.find(key));
	}

	template<PipelineKeyType T>
	typename PipelineContainer<T>::iterator PipelineContainer<T>::erase(const PipelineContainer<T>::iterator& it, const vk::Device& device, const vk::AllocationCallbacks* pAllocator) {
        device.destroyPipeline(it->second.pipeline, pAllocator);
        device.destroyPipelineLayout(it->second.layout, pAllocator);
		return pipelines.erase(it);
		
	}
	template<PipelineKeyType T>
	void PipelineContainer<T>::clear(const vk::Device& device, const vk::AllocationCallbacks* pAllocator) {
		while (pipelines.size() > 0) {
            erase(pipelines.begin(), device);
        }
	}

    struct SwapChainSupport {
        vk::SurfaceCapabilities2KHR capabilities;
        std::vector<vk::SurfaceFormat2KHR> formats;
        std::vector<vk::PresentModeKHR> presentModes;
    };


    template<typename T>
    concept SupportedIndexType = std::same_as<T, uint16_t> or std::same_as<T, uint32_t>;

    template<VertexType VertexT = Vertex3D, SupportedIndexType IndexT = uint32_t>
    struct VerticesAndIndices {
        std::vector<VertexT> vertices;
        std::vector<IndexT> indices;
        constexpr vk::IndexType getIndexType() const {
            if constexpr (std::same_as<IndexT, uint16_t>) {
                return vk::IndexType::eUint16;
            }
            else if constexpr (std::same_as<IndexT, uint32_t>) {
                return vk::IndexType::eUint32;
            }
        }
    };


    struct QueueFamilyIndices {
        std::optional<uint32_t> graphicsFamily;
        std::optional<uint32_t> presentFamily;
        std::optional<uint32_t> transferFamily;

        bool hasNecessaryValues() {
            return graphicsFamily.has_value() && presentFamily.has_value();
        }
        bool hasAllValues() {
            return graphicsFamily.has_value() && presentFamily.has_value() && transferFamily.has_value();
        }

        std::string toString() const {
            std::string s = "[ ";
            s += "graphicsFamily: ";
            if (graphicsFamily.has_value()) { s += std::to_string(graphicsFamily.value()); }
            else { s += "not set"; }
            s += ", presentFamily: ";
            if (presentFamily.has_value()) { s += std::to_string(presentFamily.value()); }
            else { s += "not set"; }
            s += ", transferFamily: ";
            if (transferFamily.has_value()) { s += std::to_string(transferFamily.value()); }
            else { s += "not set"; }
            return s + " ]";
        }
    };


    /**
     * @brief Get a vk::DependencyInfo struct for a single image memory barrier
     */
    vk::DependencyInfo getDepInfo(const vk::ImageMemoryBarrier2& barrier);

    /**
     * @brief Fill bufMeqReq and memReq 
     * @details
     *  Fill the structs and call vkGetBufferMemoryRequirements2()
     */
    /* void getBufferMemoryRequirements(const vk::Device& device, const vk::Buffer& buffer, vk::BufferMemoryRequirementsInfo2& bufMemReq, vk::MemoryRequirements2& memReq); */

    /**
     * @brief Fill imMemReq and memReq 
     * @details
     *  Fill the structs and call vkGetImageMemoryRequirements2()
     */
    /* void getImageMemoryRequirements(const vk::Device& device, const vk::Image& image, vk::ImageMemoryRequirementsInfo2& imMemReq, vk::MemoryRequirements2& memReq); */
    /// @}


//
// OBJECT USING VULKAN
//
    /**
     * @brief Base class for ObjectUsingVulkan for storage of different ObjectUsingVulkan pointers in one container
     */
    class ObjectUsingVulkanBase {
        public:
            ObjectUsingVulkanBase(std::string&& name) : objectName(name) {};
            virtual ~ObjectUsingVulkanBase() {};
            virtual void updateHandles() {};
            bool contains(const uint64_t& handle) const { return handles.contains(handle); };
            const std::string& getName() const { return objectName; };
            std::string toString() const;
        protected:
            std::string objectName;
            std::set<uint64_t> handles;
    };
    template<typename T>
    concept VulkanHppHandle = requires (const T& t) {
        static_cast<typename T::NativeType>(t);
    };
    template<typename T>
    concept PtrToVulkanHppHandle = VulkanHppHandle<std::remove_pointer_t<T>>;

    template<typename T>
    concept PtrToVulkanHppHandleRange = std::ranges::forward_range<std::remove_pointer_t<T>> 
            and VulkanHppHandle<std::ranges::range_value_t<std::remove_pointer_t<T>>>;
    template<typename T>
    concept PtrToVulkanHppHandleConvertible = PtrToVulkanHppHandle<T> || PtrToVulkanHppHandleRange<T>;
    static_assert(VulkanHppHandle<vk::RenderPass>);
    static_assert(PtrToVulkanHppHandle<vk::RenderPass*>);
    static_assert(PtrToVulkanHppHandleRange<std::vector<vk::RenderPass>*>);

    /**
     * @brief Used for debugging
     * @see VulkanInstance::debugLog
     * @details
     *  Register the handle to a vulkan object with this struct and call updateHandles().
     *  The handles set will then contain the actual address of the handle, as printed by the validation layers.
     *  That means you can check a handle returned by a validation layer 
     *  with objects of this class using ObjectUsingVulkanBase::contains() and determine the owner of the handle.
     *
     *  The pointers to the vulkan.hpp handles are stored in the handlePack tuple 
     *  (these cpp structs contain the actual vulkan handle).
     *
     *  This way, if a vulkan handle changes (eg when recreating an image) the new vulkan handle can be determined by "reading" the vulkan.hpp handle.
     *
     *  The address of the vulkan handle (as returned by validation layers) is determined by casting the vulkan.hpp handle to the corresponding vulkan handle and casting that to uint64_t:
     *  @code
     *   // example, where vkHppHandle is handle of type VkHppHandle from vulkan.hpp (eg vk::Image)
     *   uint64_t address = reinterpret_cast<uint64_t>(static_cast<typename VkHppHandle::NativeType>(vkHppHandle));
     *  @endcode
     *
     *
     *  
     * @warning 
     *  Every pointer you submit to this must stay valid!
     *  Use this struct only when debugging!
     */  
    template<PtrToVulkanHppHandleConvertible... T>
    class ObjectUsingVulkan : public ObjectUsingVulkanBase {
        public:
            ObjectUsingVulkan(std::string&& name, const T&&... pHandleConvertibles);
            /**
             * @brief Update the handles map with the current handle (aka segfault generator)
             */
            virtual void updateHandles() override;
        private:
            // Recursively call addHandles for every member of the pack
            /// Add reinterpret_cast<uint64_t>(static_cast<PHandle::NativeType>(handle)) to handles set
            template<PtrToVulkanHppHandle PHandle, PtrToVulkanHppHandleConvertible... PHandleConvertibles>
            void addHandles(const PHandle& pHandle, const PHandleConvertibles&&... pHandleConvertibles);
            template<PtrToVulkanHppHandleRange PHandleRange, PtrToVulkanHppHandleConvertible... PHandleConvertibles>
            void addHandles(const PHandleRange& pHandleRange, const PHandleConvertibles&&... pHandleConvertibles);
            /// End of recursion
            void addHandles() {}

            /// Contains pointers to (vulkan.hpp handles or ranges of vulkan.hpp handles)
            std::tuple<T...> handlePack;
    };
    // IMPLEMENTATION
    template<PtrToVulkanHppHandleConvertible... T>
	ObjectUsingVulkan<T...>::ObjectUsingVulkan(std::string&& name, const T&&... pHandleConvertibles) 
		: ObjectUsingVulkanBase(std::move(name))
	{
		handlePack = std::make_tuple<const T...>(std::forward<const T>(pHandleConvertibles)...);
		updateHandles();
	}

	template<PtrToVulkanHppHandleConvertible... T>
	void ObjectUsingVulkan<T...>::updateHandles() {
		handles.clear();
        // cant std::apply pack directly on addHandles, since that takes a this ptr
        auto f = [this](auto... args){ this->addHandles(std::forward<decltype(args)>(args)...); };
        std::apply(f, handlePack);
	}

	template<PtrToVulkanHppHandleConvertible... T>
    template<PtrToVulkanHppHandle PHandle, PtrToVulkanHppHandleConvertible... PHandleConvertibles>
    void ObjectUsingVulkan<T...>::addHandles(const PHandle& pHandle, const PHandleConvertibles&&... pHandleConvertibles) {
		handles.insert(reinterpret_cast<uint64_t>(static_cast<typename std::remove_pointer_t<PHandle>::NativeType>(*pHandle)));
        addHandles(std::forward<const PHandleConvertibles>(pHandleConvertibles)...);
	}

	template<PtrToVulkanHppHandleConvertible... T>
    template<PtrToVulkanHppHandleRange PHandleRange, PtrToVulkanHppHandleConvertible... PHandleConvertibles>
    void ObjectUsingVulkan<T...>::addHandles(const PHandleRange& pHandleRange, const PHandleConvertibles&&... pHandleConvertibles) {
		for (auto it = pHandleRange->begin(); it != pHandleRange->end(); it++) {
			handles.insert(reinterpret_cast<uint64_t>(
						static_cast<typename std::ranges::range_value_t<std::remove_pointer_t<PHandleRange>>::NativeType>(*it)));
		}
        addHandles(std::forward<const PHandleConvertibles>(pHandleConvertibles)...);
	}




    /**
     * @brief Check if all required extensions are available
     */
    template<std::ranges::forward_range RE, std::ranges::forward_range AE>
    bool areExtensionsAvailable(const RE& requiredExtensions, const AE& availableExtensions) 
        requires std::same_as<std::ranges::range_reference_t<RE>, const char*&> and std::same_as<std::ranges::range_reference_t<AE>, vk::ExtensionProperties&> 
    {
        bool extensionAvailable;
        for (const auto& requiredExtension : requiredExtensions) {
            extensionAvailable = false;
            for (const auto& extension : availableExtensions) {
                if (strcmp(requiredExtension, extension.extensionName) == 0) {
                    extensionAvailable = true;
                    break;
                }
            }
            if (!extensionAvailable) {
                return false;
            }
        }
        return true;
    }

    /**
     * @file 
     * @brief Various utility structs and functions, as well as toString() overload for vk objects
     */

} // namespace gz::vlk
