#include "vulkan_allocator.hpp"

#include "dynamic_size_block_allocator.hpp"
#include "exceptions.hpp"
#include "vulkan_instance.hpp"

#include <glm/vector_relational.hpp>
#include <gz-util/string/conversion.hpp>

namespace gz::vlk {
	DeviceMemory::DeviceMemory(vk::DeviceSize size_) 
		: size(size_), memory(VK_NULL_HANDLE), dynBlockAllocator(size_) {};


	VulkanAllocator::VulkanAllocator(VulkanInstance& instance)
		: vk(instance)
	{
        LogCreateInfo logCI {
        	.logfile = "vulkan_allocator.log",
        	.storeLog = false,
        	.prefix = "Allocator",
        	.prefixColor = Color::BLACK,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
		};
        aLog = Log(std::move(logCI));

	}

	void VulkanAllocator::allocate(const vk::MemoryAllocateInfo& allocI, const vk::MemoryRequirements2& memReq, MemoryInfo& memoryInfo) {
		aLog.log0("allocate: Requesting memory with ( size", toHexString(allocI.allocationSize), 
				"), ( memoryTypeIndex", allocI.memoryTypeIndex, 
				"), ( alignment", toHexString(memReq.memoryRequirements.alignment), ")");
		// go through allocated memories with matching memoryType
		for (auto deviceMemory = memory[allocI.memoryTypeIndex].begin(); deviceMemory != memory[allocI.memoryTypeIndex].end(); deviceMemory++) {
			try {
				uint32_t offset = deviceMemory->dynBlockAllocator.allocate(allocI.allocationSize, memReq.memoryRequirements.alignment);
				memoryInfo.memory = deviceMemory->memory;
				memoryInfo.offset = offset;
				memoryInfo.memoryTypeIndex = allocI.memoryTypeIndex;
				return;
			}
			catch (VkException& e) {
				continue;
			}			
		}

		// if nothing is found, allocate new memory
		memory[allocI.memoryTypeIndex].emplace_back(DeviceMemory(allocI.allocationSize * TODO_ALLOCATION_SIZE_MULTIPLIIER));
		vk::MemoryAllocateInfo allocI_{};
		allocI_.memoryTypeIndex = allocI.memoryTypeIndex;
		allocI_.allocationSize = memory[allocI.memoryTypeIndex].back().size;
		aLog.log0("allocate: Allocating new memory of size:", gz::toHexString(allocI_.allocationSize, 0), "(memory", memoryInfo.memoryTypeIndex, "-", gz::toString(memory[memoryInfo.memoryTypeIndex].size()) + ")");
		vk::Result result = vk.getDevice().allocateMemory(&allocI_, NO_ALLOC, &memory[allocI.memoryTypeIndex].back().memory);
		if (result != vk::Result::eSuccess) {
			throw getVkException(result, "Failed to allocate memory", "VulkanAllocator::allocate");
		}

		// return block of new memory
		DeviceMemory& deviceMemory = memory[allocI.memoryTypeIndex].back();
		try {
			uint32_t offset = deviceMemory.dynBlockAllocator.allocate(allocI.allocationSize, memReq.memoryRequirements.alignment);
			memoryInfo.memory = deviceMemory.memory;
			memoryInfo.offset = offset;
			memoryInfo.memoryTypeIndex = allocI.memoryTypeIndex;
			return;
		}
		catch (VkException& e) {
			// should never happen
			throw e;
		}
	}


	void VulkanAllocator::free(MemoryInfo& memoryInfo) {
		// go through allocated memories with matching memoryType
		for (auto deviceMemory = memory[memoryInfo.memoryTypeIndex].begin(); deviceMemory != memory[memoryInfo.memoryTypeIndex].end(); deviceMemory++) {
			if (deviceMemory->memory != memoryInfo.memory) { continue; }
			try {
				deviceMemory->dynBlockAllocator.free(memoryInfo.offset);
			}
			catch (VkUserError&) {
				throw VkUserError("Trying to free block that was not allocated from this VulkanAllocator", "VulkanAllocator::free");
			}
			memoryInfo.memory = VK_NULL_HANDLE;
			memoryInfo.offset = 0; 
				
			/* aLog.log0("free: (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") Blocks:", deviceMemory->blocks); */
			// if now there is only one free block, everything is free -> deallocate memory
			if (deviceMemory->dynBlockAllocator.blockCount() == 1) {
				aLog.log0("free: Deallocating (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") of size:", gz::toHexString(deviceMemory->size, 0));
				vk.getDevice().freeMemory(deviceMemory->memory, NO_ALLOC);
				memory[memoryInfo.memoryTypeIndex].erase(deviceMemory);
			}
			return;
		}
	}


	void VulkanAllocator::cleanup() {
		aLog.log0("cleanup: cleaning up");
		for (auto memType = memory.begin(); memType != memory.end(); memType++) {
			for (auto deviceMemory = memType->second.begin(); deviceMemory != memType->second.end(); deviceMemory++) {
				// check if all blocks are free
				for (auto block = deviceMemory->dynBlockAllocator.begin(); block != deviceMemory->dynBlockAllocator.end(); block++) {
					if (block->free) { continue; }
					aLog.warning("cleanup: (memory", memType->first, "-", gz::toString(deviceMemory - memType->second.begin()) + ") Block not freed: ", *block);
				}
				aLog.log0("free: Deallocating (memory", memType->first, "-", gz::toString(deviceMemory - memType->second.begin()) + ") of size:", gz::toHexString(deviceMemory->size, 0));
				vk.getDevice().freeMemory(deviceMemory->memory, NO_ALLOC);
			}
		}
		memory.clear();
	}


}  // namespace gz::vlk
